// Copyright (c) 2019, Dr. Bruno Guigas. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.
import 'dart:typed_data';
import 'dart:math' as math;

/// This class provides functionalities to compute an n-dimensional mixed
/// Lorentz-Gauss line shape, also called an n-dimensional Pseudo-Voigt
/// function V(x) (see https://de.wikipedia.org/wiki/Voigt-Profil):
///
/// V(x) = m*L(x) + (1-m)*G(x), m = 0...1
/// where L is a Lorentzian and G a Gaussian line shape.
/// In detail:
/// V(x1, ... , xn)) = Product(i=1,n){
///  a * (m[i]) / ( (1+ FLORENTZ*((x[i]-c[i])/w[i])^2) ) +
///  a * (1-m[i])*exp(-FGAUSS*((x[i]-c[i])/w[i])^2)}
class LorentzGauss extends LineShapeND {
  /// defaults for [fGauss] and [fLorentz]
  final double FGAUSS = 4 * math.log(2), FLORENTZ = 4.0;

  /// centers, widths, L/G mixing factors
  List<double> c, w, m;
  // amplitude
  double a;

  /// combination of parameters above in an array
  List<double> pars;

  /// line shape dimension
  int dim = 0;

  /// factors for shape consistency (see constructor comment)
  double fGauss, fLorentz;

  /// Constructs a mixed Gauss-Lorentz shape defined by:
  /// [a] - amplitude
  /// [c] - center
  /// [w] - width
  /// [m] - mixing factors (0 <= m <= 1) where
  ///   m=1=pure Lorentzian, then w = Lorentzian width
  ///   m=0=pure Gaussian, then w = Gaussian sigma
  /// The dimension n of c,w,m must be the same (>0) and define the dimension of
  /// the shape. Otherwise an exception is thrown.
  /// [fGauss] and [fLorentz] will take on defaults (see above) when omitted.
  /// Different literature references are using different defaults. The factors
  /// ensure that [w] is the line width at half maximum height for L, G, and mixed.
  LorentzGauss.fromPars(double this.a, List<double> this.c, List<double> this.w,
      List<double> this.m,
      [double this.fGauss, double this.fLorentz]) {
    if (fGauss == null) fGauss = FGAUSS;
    if (fLorentz == null) fLorentz = FLORENTZ;
    dim = c.length;
    if (dim != w.length || dim != m.length) {
      throw "List size error\n. Expected: $dim";
    }

    pars = new List<double>(1 + c.length + w.length + m.length);
    pars[0] = a;
    for (int i = 0; i < dim; i++) {
      pars[i + 1] = c[i];
      pars[i + 1 + dim] = w[i];
      pars[i + 1 + dim + dim] = m[i];
    }
    shapeParameters = pars;
  }

  /// Returns the value of this mixed Gauss-Lorentz shape
  /// at the n-dimensional point [x].
  /// Throws Exception if the dimension of [x] is not in accordance with
  /// the parameters specified at shape construction time.
  double getValueAt(List<double> x) {
    if (dim != x.length) {
      throw "Illegal dimension. Should be $dim";
    }

    double voigt = 1.0;
    double lorentz = 0.0, gauss = 0.0, factor = 0.0;
    for (int i = 0; i < dim; i++) {
      factor = (x[i] - c[i]) / (1.0 * w[i]); // stored widths are half-widths
      factor *= factor;

      lorentz = m[i] / (1 + fLorentz * factor);
      // speed up: only compute exp() if no pure Lorentzian (0 <= m[i] <= 1
      if ((1 - m[i]) > 0.001) {
        gauss = (1 - m[i]) * math.exp(-fGauss * factor);
        voigt *= lorentz + gauss;
      } else {
        voigt *= lorentz;
      }
    }
    voigt *= a;
    return voigt;
  }
}

/// Defines an n-dimensional abstract shape.
abstract class LineShapeND {
  Object shapeParameters; // e.g. List<double>, ...
  double integral;
  bool integralCalculated = false;
  final String NOT_IMPL = "Method not implemented";

  /// pars - the parameters of the shape. Can be anything.
  /// For Gauss-Lorentz, e.g. a List<double> (see there)
  fromList(Object pars) {
    shapeParameters = pars;
  }

  /// Calculates and returns the values of the Shape at the given position [x].
  /// [x].length is the dimension of the shape, so a position has n coordinate values.
  double getValueAt(List<double> x);

  /// Calculates and returns the values of the Shape at the given positions [xxx].
  /// xxx.length is the dimension of the shape, so a position has n coordinate values.
  ///  xxx[0].length is the number of points for which to calculate the shape value.
  ///  xxx[n][p]  the n-dimensional coordinate of a point p
  /// [integralType] - how the integral is computed. 0=sum of all values.
  ///   -1 = compute no integral
  Float64List getValuesAt(List<List<double>> xxx, int integralType) {
    int dim = xxx.length;
    int npoints = xxx[0].length;
    Float64List values = new Float64List(npoints);
    List<double> curpoint = new List<double>(dim);

    integral = 0.0;
    for (int i = 0; i < npoints; i++) {
      for (int k = 0; k < dim; k++) {
        curpoint[k] = xxx[k][i];
      }

      values[i] = getValueAt(curpoint);

      if (integralType == 0) {
        integral += values[i];
      }
    }

    integralCalculated = false;
    if (integralType != -1) {
      integralCalculated = true;
    }
    return values;
  }

  /// Returns the integral of this shape if its was computed (see [getValuesAt],
  /// otherwise throws.
  double getIntegral() {
    if (!integralCalculated) {
      throw "Integral not available.\nCall getValuesAt(..) first.";
    }
    return integral;
  }

  /// Returns teh shape parameters
  Object getShapeParams() {
    return shapeParameters;
  }

  /// Returns an estimate of the width at half height of the peak at [peakindex]
  /// of the data given by [array].
  /// This assumes a Gauss/Lorentz type line shape.
  /// Typically used if [array] contains experimental data of Gauss/Lorentz type.
  static double getWidth(Float64List array, int peakindex) {
    double peakheight = array[peakindex];
    double halfheight = peakheight / 2;
    int leftpos = peakindex - 1, rightpos = peakindex + 1;

    // move down at the left of the peak
    while (true) {
      if (!(leftpos >= 1 && leftpos < array.length)) break;
      if (array[leftpos] < halfheight) break;
      if (array[leftpos - 1] > array[leftpos]) break; // raises again
      leftpos--;
    }

    // move down at the right of the peak
    while (true) {
      if (!(rightpos >= 0 && rightpos < array.length - 1)) break;
      if (array[rightpos] < halfheight) break;
      if (array[rightpos + 1] > array[rightpos]) break; // raises again
      rightpos++;
    }

    double result = (rightpos - leftpos).toDouble().abs();
    return result;
  }
}
